@@ -599,6 +599,46 @@ static int net_tls_client(lua_State *L)
 	return 1;
 }
 
+static int net_tls_client_clear(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	if (!engine) {
+		return 0;
+	}
+
+	struct network *net = &engine->net;
+	if (!net) {
+		return 0;
+	}
+
+	if (lua_gettop(L) != 1 || !lua_isstring(L, 1)) {
+		format_error(L, "net.tls_client_clear() requires one parameter (\"address\")");
+		lua_error(L);
+	}
+
+	const char *full_addr = lua_tostring(L, 1);
+
+	char addr[INET6_ADDRSTRLEN];
+	uint16_t port = 0;
+	if (kr_straddr_split(full_addr, addr, sizeof(addr), &port) != kr_ok()) {
+		format_error(L, "invalid IP address");
+		lua_error(L);
+	}
+
+	if (port == 0) {
+		port = 853;
+	}
+
+	int r = tls_client_params_clear(&net->tls_client_params, addr, port);
+	if (r != 0) {
+		lua_pushstring(L, kr_strerror(r));
+		lua_error(L);
+	}
+
+	lua_pushboolean(L, true);
+	return 1;
+}
+
 static int net_tls_padding(lua_State *L)
 {
 	struct engine *engine = engine_luaget(L);
@@ -793,37 +833,119 @@ static int net_outgoing(lua_State *L, int family)
 static int net_outgoing_v4(lua_State *L) { return net_outgoing(L, AF_INET); }
 static int net_outgoing_v6(lua_State *L) { return net_outgoing(L, AF_INET6); }
 
-static int net_tcp_in_idle(lua_State *L)
+static int net_update_timeout(lua_State *L, uint64_t *timeout, const char *name)
 {
-	struct engine *engine = engine_luaget(L);
-	struct network *net = &engine->net;
-
 	/* Only return current idle timeout. */
 	if (lua_gettop(L) == 0) {
-		lua_pushnumber(L, net->tcp.in_idle_timeout);
+		lua_pushnumber(L, *timeout);
 		return 1;
 	}
 
 	if ((lua_gettop(L) != 1)) {
-		lua_pushstring(L, "net.tcp_in_idle takes one parameter: (\"idle timeout\")");
+		lua_pushstring(L, name);
+		lua_pushstring(L, " takes one parameter: (\"idle timeout\")");
 		lua_error(L);
 	}
 
 	if (lua_isnumber(L, 1)) {
 		int idle_timeout = lua_tointeger(L, 1);
 		if (idle_timeout <= 0) {
-			lua_pushstring(L, "net.tcp_in_idle parameter has to be positive number");
+			lua_pushstring(L, name);
+			lua_pushstring(L, " parameter has to be positive number");
 			lua_error(L);
 		}
-		net->tcp.in_idle_timeout = idle_timeout;
+		*timeout = idle_timeout;
 	} else {
-		lua_pushstring(L, "net.tcp_in_idle parameter has to be positive number");
+		lua_pushstring(L, name);
+		lua_pushstring(L, " parameter has to be positive number");
 		lua_error(L);
 	}
 	lua_pushboolean(L, true);
 	return 1;
 }
 
+static int net_tcp_in_idle(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct network *net = &engine->net;
+
+	return net_update_timeout(L, &net->tcp.in_idle_timeout, "net.tcp_in_idle");
+}
+
+static int net_tls_handshake_timeout(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct network *net = &engine->net;
+
+	return net_update_timeout(L, &net->tcp.tls_handshake_timeout, "net.tls_handshake_timeout");
+}
+
+static int net_bpf_set(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct network *net = &engine->net;
+
+	if (lua_gettop(L) != 1 || !lua_isnumber(L, 1)) {
+		format_error(L, "net.bpf_set(fd) takes one parameter: the open file descriptor of a loaded BPF program");
+		lua_error(L);
+		return 0;
+	}
+
+#if __linux__
+
+	int progfd = lua_tointeger(L, 1);
+	if (progfd == 0) {
+		/* conversion error despite that fact
+		 * that lua_isnumber(L, 1) has returned true.
+		 * Real or stdin? */
+		lua_error(L);
+		return 0;
+	}
+	lua_pop(L, 1);
+
+	if (network_set_bpf(net, progfd) == 0) {
+		char errmsg[256] = {};
+		snprintf(errmsg, sizeof(errmsg), "failed to attach BPF program to some networks: %s", strerror(errno));
+		format_error(L, errmsg);
+		lua_error(L);
+		return 0;
+	}
+
+	lua_pushboolean(L, 1);
+	return 1;
+
+#endif
+
+	format_error(L, "BPF is not supported on this operating system");
+	lua_error(L);
+	return 0;
+}
+
+static int net_bpf_clear(lua_State *L)
+{
+	struct engine *engine = engine_luaget(L);
+	struct network *net = &engine->net;
+
+	if (lua_gettop(L) != 0) {
+		format_error(L, "net.bpf_clear() does not take any parameters");
+		lua_error(L);
+		return 0;
+	}
+
+#if __linux__
+
+	network_clear_bpf(net);
+
+	lua_pushboolean(L, 1);
+	return 1;
+
+#endif
+
+	format_error(L, "BPF is not supported on this operating system");
+	lua_error(L);
+	return 0;
+}
+
 int lib_net(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
@@ -836,12 +958,16 @@ int lib_net(lua_State *L)
 		{ "tls",          net_tls },
 		{ "tls_server",   net_tls },
 		{ "tls_client",   net_tls_client },
+		{ "tls_client_clear", net_tls_client_clear },
 		{ "tls_padding",  net_tls_padding },
 		{ "tls_sticket_secret", net_tls_sticket_secret_string },
 		{ "tls_sticket_secret_file", net_tls_sticket_secret_file },
 		{ "outgoing_v4",  net_outgoing_v4 },
 		{ "outgoing_v6",  net_outgoing_v6 },
 		{ "tcp_in_idle",  net_tcp_in_idle },
+		{ "tls_handshake_timeout",  net_tls_handshake_timeout },
+		{ "bpf_set",      net_bpf_set },
+		{ "bpf_clear",    net_bpf_clear },
 		{ NULL, NULL }
 	};
 	register_lib(L, "net", lib);
@@ -1205,7 +1331,7 @@ static int cache_get(lua_State *L)
 }
 
 /** Set time interval for cleaning rtt cache.
- * Servers with score >= KR_NS_TIMEOUTED will be cleaned after
+ * Servers with score >= KR_NS_TIMEOUT will be cleaned after
  * this interval ended up, so that they will be able to participate
  * in NS elections again. */
 static int cache_ns_tout(lua_State *L)
@@ -1225,8 +1351,8 @@ static int cache_ns_tout(lua_State *L)
 		lua_error(L);
 	}
 
-	lua_Number interval_lua = lua_tonumber(L, 1);
-	if (!(interval_lua >= 0 && interval_lua < UINT_MAX)) {
+	lua_Integer interval_lua = lua_tointeger(L, 1);
+	if (!(interval_lua > 0 && interval_lua < UINT_MAX)) {
 		format_error(L, "invalid interval specified, it must be in range > 0, < " xstr(UINT_MAX));
 		lua_error(L);
 	}
