@@ -15,21 +15,23 @@
  */
 
 #include <arpa/inet.h>
+#include <assert.h>
+#include <getopt.h>
+#include <libgen.h>
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
-#include <getopt.h>
-#include <libgen.h>
+#include <unistd.h>
+
 #include <uv.h>
-#include <assert.h>
-#include <contrib/cleanup.h>
-#include <contrib/ucw/mempool.h>
-#include <contrib/ccan/asprintf/asprintf.h>
-#include <libknot/error.h>
 #ifdef HAS_SYSTEMD
 #include <systemd/sd-daemon.h>
 #endif
+#include <libknot/error.h>
 
+#include <contrib/cleanup.h>
+#include <contrib/ucw/mempool.h>
+#include <contrib/ccan/asprintf/asprintf.h>
 #include "lib/defines.h"
 #include "lib/resolve.h"
 #include "lib/dnssec.h"
@@ -300,6 +302,42 @@ static void signal_handler(uv_signal_t *handle, int signum)
 	uv_signal_stop(handle);
 }
 
+/** SIGBUS -> attempt to remove the overflowing cache file and abort. */
+static void sigbus_handler(int sig, siginfo_t *siginfo, void *ptr)
+{
+	/* We can't safely assume that printf-like functions work, but write() is OK.
+	 * See POSIX for the safe functions, e.g. 2017 version just above this link:
+	 * http://pubs.opengroup.org/onlinepubs/9699919799/functions/V2_chap02.html#tag_15_04_04
+	 */
+	#define WRITE_ERR(err_charray) \
+		(void)write(STDERR_FILENO, err_charray, sizeof(err_charray))
+	/* Unfortunately, void-cast on the write isn't enough to avoid the warning. */
+	#pragma GCC diagnostic push
+	#pragma GCC diagnostic ignored "-Wunused-result"
+	const char msg_typical[] =
+		"\nSIGBUS received; this is most likely due to filling up the filesystem where cache resides.\n",
+		msg_unknown[] = "\nSIGBUS received, cause unknown.\n",
+		msg_deleted[] = "Cache file deleted.\n",
+		msg_del_fail[] = "Cache file deletion failed.\n",
+		msg_final[] = "kresd can not recover reliably by itself, exiting.\n";
+	if (siginfo->si_code != BUS_ADRERR) {
+		WRITE_ERR(msg_unknown);
+		goto end;
+	}
+	WRITE_ERR(msg_typical);
+	if (!kr_cache_emergency_file_to_remove) goto end;
+	if (unlink(kr_cache_emergency_file_to_remove)) {
+		WRITE_ERR(msg_del_fail);
+	} else {
+		WRITE_ERR(msg_deleted);
+	}
+end:
+	WRITE_ERR(msg_final);
+	_exit(128 - sig); /*< regular return from OS-raised SIGBUS can't work anyway */
+	#undef WRITE_ERR
+	#pragma GCC diagnostic pop
+}
+
 /** Split away port from the address. */
 static const char *set_addr(char *addr, int *port)
 {
@@ -376,8 +414,24 @@ static void help(int argc, char *argv[])
 	       " [rundir]             Path to the working directory (default: .)\n");
 }
 
+/** \return exit code for main()  */
 static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_set, bool leader, struct args *args)
 {
+	/* Only some kinds of stdin work with uv_pipe_t.
+	 * Otherwise we would abort() from libuv e.g. with </dev/null */
+	if (args->interactive) switch (uv_guess_handle(0)) {
+	case UV_TTY:		/* standard terminal */
+	case UV_NAMED_PIPE:	/* echo 'quit()' | kresd ... */
+		break;
+	default:
+		kr_log_error(
+			"[system] error: standard input is not a terminal or pipe; "
+			"use '-f 1' if you want non-interactive mode.  "
+			"Commands can be simply added to your configuration file or sent over the tty/$PID control socket.\n"
+			);
+		return 1;
+	}
+
 	/* Control sockets or TTY */
 	auto_free char *sock_file = NULL;
 	uv_pipe_t pipe;
@@ -423,7 +477,8 @@ static int run_worker(uv_loop_t *loop, struct engine *engine, fd_array_t *ipc_se
 	if (sock_file) {
 		unlink(sock_file);
 	}
-	return kr_ok();
+	uv_close((uv_handle_t *)&pipe, NULL); /* Seems OK even on the stopped loop. */
+	return 0;
 }
 
 #ifdef HAS_SYSTEMD
@@ -475,6 +530,14 @@ static void args_init(struct args *args)
 	args->quiet = false;
 }
 
+static long strtol_10(const char *s)
+{
+	if (!s) abort();
+	/* ^^ This shouldn't ever happen.  When getopt_long() returns an option
+	 * character that has a mandatory parameter, optarg can't be NULL. */
+	return strtol(s, NULL, 10);
+}
+
 /** Process arguments into struct args.
  * @return >=0 if main() should be exited immediately.
  */
@@ -508,17 +571,17 @@ static int parse_args(int argc, char **argv, struct args *args)
 			array_push(args->tls_set, optarg);
 			break;
 		case 'S':
-			array_push(args->fd_set, strtol(optarg, NULL, 10));
+			array_push(args->fd_set, strtol_10(optarg));
 			break;
 		case 'T':
-			array_push(args->tls_fd_set, strtol(optarg, NULL, 10));
+			array_push(args->tls_fd_set, strtol_10(optarg));
 			break;
 		case 'c':
 			args->config = optarg;
 			break;
 		case 'f':
 			args->interactive = false;
-			args->forks = strtol(optarg, NULL, 10);
+			args->forks = strtol_10(optarg);
 			if (args->forks <= 0) {
 				kr_log_error("[system] error '-f' requires a positive"
 						" number, not '%s'\n", optarg);
@@ -689,7 +752,36 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
-	uv_loop_t *loop = NULL;
+	uv_loop_t *loop = uv_default_loop();
+	worker->loop = loop;
+	loop->data = worker;
+
+	/* Catch some signals. */
+	uv_signal_t sigint, sigterm;
+	if (true) ret = uv_signal_init(loop, &sigint);
+	if (!ret) ret = uv_signal_init(loop, &sigterm);
+	if (!ret) ret = uv_signal_start(&sigint, signal_handler, SIGINT);
+	if (!ret) ret = uv_signal_start(&sigterm, signal_handler, SIGTERM);
+	/* Block SIGPIPE; see https://github.com/libuv/libuv/issues/45 */
+	if (!ret && signal(SIGPIPE, SIG_IGN) == SIG_ERR) ret = errno;
+	if (!ret) {
+		/* Catching SIGBUS via uv_signal_* can't work; see:
+		 * https://github.com/libuv/libuv/pull/1987 */
+		struct sigaction sa;
+		memset(&sa, 0, sizeof(sa));
+		sa.sa_sigaction = sigbus_handler;
+		sa.sa_flags = SA_SIGINFO;
+		if (sigaction(SIGBUS, &sa, NULL)) {
+			ret = errno;
+		}
+	}
+	if (ret) {
+		kr_log_error("[system] failed to set up signal handlers: %s\n",
+				strerror(abs(errno)));
+		ret = EXIT_FAILURE;
+		goto cleanup;
+	}
+
 	/* Bind to passed fds and sockets*/
 	if (bind_fds(&engine.net, &args.fd_set, false) != 0 ||
 	    bind_fds(&engine.net, &args.tls_fd_set, true) != 0 ||
@@ -700,30 +792,8 @@ int main(int argc, char **argv)
 		goto cleanup;
 	}
 
-	/* Workaround for https://github.com/libuv/libuv/issues/45
-	 * (Write after ECONNRESET crash.) */
-	if (ret == 0 && signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
-		kr_log_error("[system] can't block SIGPIPE signal: %s\n",
-				strerror(errno));
-		ret = EXIT_FAILURE;
-	}
-
-	if (ret != 0) {
-		goto cleanup;
-	}
-
-	engine_set_moduledir(&engine, args.moduledir);
-	
-	/* Block signals. */
-	loop = uv_default_loop();
-	uv_signal_t sigint, sigterm;
-	uv_signal_init(loop, &sigint);
-	uv_signal_init(loop, &sigterm);
-	uv_signal_start(&sigint, signal_handler, SIGINT);
-	uv_signal_start(&sigterm, signal_handler, SIGTERM);
 	/* Start the scripting engine */
-	worker->loop = loop;
-	loop->data = worker;
+	engine_set_moduledir(&engine, args.moduledir);
 
 	if (engine_load_sandbox(&engine) != 0) {
 		ret = EXIT_FAILURE;
@@ -753,11 +823,6 @@ int main(int argc, char **argv)
 
 	/* Run the event loop */
 	ret = run_worker(loop, &engine, &ipc_set, fork_id == 0, &args);
-	if (ret != 0) {
-		perror("[system] worker failed");
-		ret = EXIT_FAILURE;
-		goto cleanup;
-	}
 
 cleanup:/* Cleanup. */
 	engine_deinit(&engine);
