@@ -58,7 +58,7 @@
 #include "lib/generic/map.h"
 #include "lib/generic/array.h"
 
-#define VERBOSE_MSG(qry, fmt...) QRVERBOSE(qry, "zimport", fmt)
+#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "zimport", __VA_ARGS__)
 
 /* Pause between parse and import stages, milliseconds.
  * See comment in zi_zone_import() */
@@ -314,7 +314,7 @@ static knot_pkt_t *zi_query_create(zone_import_ctx_t *z_import, knot_rrset_t *rr
 {
 	knot_mm_t *pool = &z_import->pool;
 
-	uint32_t msgid = kr_rand_uint(0);
+	uint32_t msgid = kr_rand_bytes(2);
 
 	knot_pkt_t *query = knot_pkt_new(NULL, KNOT_WIRE_MAX_PKTSIZE, pool);
 	if (!query) {
