@@ -36,7 +36,7 @@
 #include "lib/defines.h"
 #include "lib/module.h"
 
-#define VERBOSE_MSG(qry, fmt...) QRVERBOSE(qry, "vldr", fmt)
+#define VERBOSE_MSG(qry, ...) QRVERBOSE(qry, "vldr", __VA_ARGS__)
 
 #define MAX_REVALIDATION_CNT 2
 
@@ -916,7 +916,7 @@ static int validate(kr_layer_t *ctx, knot_pkt_t *pkt)
 	}
 
 	/* Pass-through if CD bit is set. */
-	if (knot_wire_get_cd(req->answer->wire)) {
+	if (knot_wire_get_cd(req->qsource.packet->wire)) {
 		check_wildcard(ctx);
 		wildcard_adjust_to_wire(req, qry);
 		rank_records(ctx, KR_RANK_OMIT, NULL);
