@@ -22,7 +22,6 @@
 #include <sys/time.h>
 #include <contrib/cleanup.h>
 #include <contrib/ccan/asprintf/asprintf.h>
-#include <ccan/isaac/isaac.h>
 #include <ucw/mempool.h>
 #include <gnutls/gnutls.h>
 #include <libknot/descriptor.h>
@@ -48,11 +47,24 @@
 /* Logging & debugging */
 bool kr_verbose_status = false;
 
-/** @internal CSPRNG context */
-static isaac_ctx ISAAC;
-static bool isaac_seeded = false;
-#define SEED_SIZE 256
-
+void *mm_realloc(knot_mm_t *mm, void *what, size_t size, size_t prev_size)
+{
+	if (mm) {
+		void *p = mm->alloc(mm->ctx, size);
+		if (p == NULL) {
+			return NULL;
+		} else {
+			if (what) {
+				memcpy(p, what,
+				       prev_size < size ? prev_size : size);
+			}
+			mm_free(mm, what);
+			return p;
+		}
+	} else {
+		return realloc(what, size);
+	}
+}
 
 void *mm_malloc(void *ctx, size_t n)
 {
@@ -115,6 +127,23 @@ void kr_log_verbose(const char *fmt, ...)
 	}
 }
 
+void kr_log_qverbose_impl(const struct kr_query *qry, const char *cls, const char *fmt, ...)
+{
+	unsigned ind = 0;
+	for (const struct kr_query *q = qry; q; q = q->parent)
+		ind += 2;
+	uint32_t qry_uid = qry ? qry->uid : 0;
+	uint32_t req_uid = qry && qry->request ? qry->request->uid : 0;
+	/* Simplified kr_log_verbose() calls, first prefix then passed fmt...
+	 * Calling it would take about the same amount of code. */
+	printf("[%05u.%02u][%s] %*s", req_uid, qry_uid, cls, ind, "");
+	va_list args;
+	va_start(args, fmt);
+	vprintf(fmt, args);
+	va_end(args);
+	fflush(stdout);
+}
+
 bool kr_log_trace(const struct kr_query *query, const char *source, const char *fmt, ...)
 {
 	if (!kr_log_trace_enabled(query)) {
@@ -181,67 +210,6 @@ char* kr_strcatdup(unsigned n, ...)
 	return result;
 }
 
-static int seed_file(const char *fname, char *buf, size_t buflen)
-{
-	auto_fclose FILE *fp = fopen(fname, "r");
-	if (!fp) {
-		return kr_error(EINVAL);
-	}
-	/* Disable buffering to conserve randomness but ignore failing to do so. */
-	setvbuf(fp, NULL, _IONBF, 0);
-	do {
-		if (feof(fp)) {
-			return kr_error(ENOENT);
-		}
-		if (ferror(fp)) {
-			return kr_error(ferror(fp));
-		}
-		if (fread(buf, buflen, 1, fp) == 1) { /* read in one chunk for simplicity */
-			return kr_ok();
-		}
-	} while (true);
-	return 0;
-}
-
-static int randseed(char *buf, size_t buflen)
-{
-    /* This is adapted from Tor's crypto_seed_rng() */
-    static const char *filenames[] = {
-        "/dev/srandom", "/dev/urandom", "/dev/random", NULL
-    };
-    for (unsigned i = 0; filenames[i]; ++i) {
-        if (seed_file(filenames[i], buf, buflen) == 0) {
-            return 0;
-        }
-    }
-
-    /* Seed from time, this is not going to be secure. */
-    kr_log_error("failed to obtain randomness, falling back to current time\n");
-    struct timeval tv;
-    gettimeofday(&tv, NULL);
-    memcpy(buf, &tv, buflen < sizeof(tv) ? buflen : sizeof(tv));
-    return 0;
-}
-
-int kr_rand_reseed(void)
-{
-	uint8_t seed[SEED_SIZE];
-	randseed((char *)seed, sizeof(seed));
-	isaac_reseed(&ISAAC, seed, sizeof(seed));
-	return kr_ok();
-}
-
-uint32_t kr_rand_uint(uint32_t max)
-{
-	if (unlikely(!isaac_seeded)) {
-		kr_rand_reseed();
-		isaac_seeded = true;
-	}
-	return max == 0
-		? isaac_next_uint32(&ISAAC)
-		: isaac_next_uint(&ISAAC, max);
-}
-
 int kr_memreserve(void *baton, char **mem, size_t elm_size, size_t want, size_t *have)
 {
     if (*have >= want) {
@@ -303,18 +271,19 @@ int kr_pkt_clear_payload(knot_pkt_t *pkt)
 int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
                uint16_t rclass, uint16_t rtype, const uint8_t *rdata, uint16_t rdlen)
 {
+	/* LATER(opt.): there's relatively lots of copying, but ATM kr_pkt_put()
+	 * isn't considered to be used in any performance-critical parts (just lua). */
 	if (!pkt || !name)  {
 		return kr_error(EINVAL);
 	}
 	/* Create empty RR */
 	knot_rrset_t rr;
 	knot_rrset_init(&rr, knot_dname_copy(name, &pkt->mm), rtype, rclass, ttl);
-	/* Create RDATA
-	 * @warning _NOT_ thread safe.
-	 */
-	static knot_rdata_t rdata_arr[RDATA_ARR_MAX];
-	knot_rdata_init(rdata_arr, rdlen, rdata);
-	knot_rdataset_add(&rr.rrs, rdata_arr, &pkt->mm);
+	/* Create RDATA */
+	knot_rdata_t *rdata_tmp = mm_alloc(&pkt->mm, offsetof(knot_rdata_t, data) + rdlen);
+	knot_rdata_init(rdata_tmp, rdlen, rdata);
+	knot_rdataset_add(&rr.rrs, rdata_tmp, &pkt->mm);
+	mm_free(&pkt->mm, rdata_tmp); /* we're always on mempool for now, but whatever */
 	/* Append RR */
 	return knot_pkt_put(pkt, 0, &rr, KNOT_PF_FREE);
 }
@@ -410,29 +379,38 @@ uint16_t kr_inaddr_port(const struct sockaddr *addr)
 	}
 }
 
+void kr_inaddr_set_port(struct sockaddr *addr, uint16_t port)
+{
+	if (!addr) {
+		return;
+	}
+	switch (addr->sa_family) {
+	case AF_INET:  ((struct sockaddr_in *)addr)->sin_port = htons(port);
+	case AF_INET6: ((struct sockaddr_in6 *)addr)->sin6_port = htons(port);
+	default: break;
+	}
+}
+
 int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen)
 {
-	int ret = kr_ok();
 	if (!addr || !buf || !buflen) {
 		return kr_error(EINVAL);
 	}
 
-	char str[INET6_ADDRSTRLEN + 6];
-	if (!inet_ntop(addr->sa_family, kr_inaddr(addr), str, sizeof(str))) {
+	if (!inet_ntop(addr->sa_family, kr_inaddr(addr), buf, *buflen)) {
 		return kr_error(errno);
 	}
-	int len = strlen(str);
-	str[len] = '#';
-	u16tostr((uint8_t *)&str[len + 1], kr_inaddr_port(addr));
-	len += 6;
-	str[len] = 0;
-	if (len >= *buflen) {
-		ret = kr_error(ENOSPC);
-	} else {
-		memcpy(buf, str, len + 1);
+	const int len = strlen(buf);
+	const int len_need = len + 1 + 5 + 1;
+	if (len_need > *buflen) {
+		*buflen = len_need;
+		return kr_error(ENOSPC);
 	}
-	*buflen = len;
-	return ret;
+	*buflen = len_need;
+	buf[len] = '#';
+	u16tostr((uint8_t *)&buf[len + 1], kr_inaddr_port(addr));
+	buf[len_need - 1] = 0;
+	return kr_ok();
 }
 
 int kr_straddr_family(const char *addr)
@@ -1011,6 +989,40 @@ finish:
 	return d - dst;
 }
 
+static void rnd_noerror(void *data, uint size)
+{
+	int ret = gnutls_rnd(GNUTLS_RND_NONCE, data, size);
+	if (ret) {
+		kr_log_error("gnutls_rnd(): %s\n", gnutls_strerror(ret));
+		abort();
+	}
+}
+void kr_rnd_buffered(void *data, uint size)
+{
+	/* static circular buffer, from index _begin (inclusive) to _end (exclusive) */
+	static uint8_t buf[512/8]; /* gnutls_rnd() works on blocks of 512 bits (chacha) */
+	static uint buf_begin = sizeof(buf);
+
+	if (unlikely(size > sizeof(buf))) {
+		rnd_noerror(data, size);
+		return;
+	}
+	/* Start with contiguous chunk, possibly until the end of buffer. */
+	const uint size1 = MIN(size, sizeof(buf) - buf_begin);
+	uint8_t *d = data;
+	memcpy(d, buf + buf_begin, size1);
+	if (size1 == size) {
+		buf_begin += size1;
+		return;
+	}
+	d += size1;
+	size -= size1;
+	/* Refill the whole buffer, and finish by another contiguous chunk. */
+	rnd_noerror(buf, sizeof(buf));
+	memcpy(d, buf, size);
+	buf_begin = size;
+}
+
 void kr_rrset_init(knot_rrset_t *rrset, knot_dname_t *owner,
 			uint16_t type, uint16_t rclass, uint32_t ttl)
 {
@@ -1037,3 +1049,4 @@ uint16_t kr_rrsig_type_covered(const knot_rdata_t *rdata)
 {
 	return knot_rrsig_type_covered(rdata);
 }
+
