@@ -22,11 +22,16 @@
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <netinet/in.h>
+
+#include <gnutls/gnutls.h>
+#include <gnutls/crypto.h>
+#include <lua.h>
+
 #include <libknot/libknot.h>
 #include <libknot/packet/pkt.h>
 #include <libknot/rrset.h>
 #include <libknot/rrtype/rrsig.h>
-#include <lua.h>
+
 #include "lib/generic/array.h"
 #include "lib/defines.h"
 
@@ -42,8 +47,8 @@ typedef void (*trace_callback_f)(struct kr_request *request);
 /** @brief Callback for request logging handler. */
 typedef void (*trace_log_f)(const struct kr_query *query, const char *source, const char *msg);
 
-#define kr_log_info(fmt, ...) do { printf((fmt), ## __VA_ARGS__); fflush(stdout); } while(0)
-#define kr_log_error(fmt, ...) fprintf(stderr, (fmt), ## __VA_ARGS__)
+#define kr_log_info(...) do { printf(__VA_ARGS__); fflush(stdout); } while(0)
+#define kr_log_error(...) fprintf(stderr, ## __VA_ARGS__)
 
 /* Always export these, but override direct calls by macros conditionally. */
 /** Whether in --verbose mode.  Only use this for reading. */
@@ -56,10 +61,16 @@ KR_EXPORT bool kr_verbose_set(bool status);
 KR_EXPORT KR_PRINTF(1)
 void kr_log_verbose(const char *fmt, ...);
 
+/** Utility for QRVERBOSE - use that instead. */
+KR_EXPORT KR_PRINTF(3)
+void kr_log_qverbose_impl(const struct kr_query *qry, const char *cls, const char *fmt, ...);
+
 /**
  * @brief Return true if the query has request log handler installed.
  */
-#define kr_log_trace_enabled(query) ((query) && (query)->request && (query)->request->trace_log)
+#define kr_log_trace_enabled(query) (__builtin_expect( \
+	(query) && (query)->request && (query)->request->trace_log, \
+	false))
 
 /**
  * Log a message through the request log handler.
@@ -103,33 +114,19 @@ static inline void *mm_alloc(knot_mm_t *mm, size_t size)
 	if (mm) return mm->alloc(mm->ctx, size);
 	else return malloc(size);
 }
-static inline void mm_free(knot_mm_t *mm, void *what)
+static inline void mm_free(knot_mm_t *mm, const void *what)
 {
 	if (mm) {
 		if (mm->free)
-			mm->free(what);
-	}
-	else free(what);
-}
-static inline void *mm_realloc(knot_mm_t *mm, void *what, size_t size, size_t prev_size)
-{
-	if (mm) {
-		void *p = mm->alloc(mm->ctx, size);
-		if (p == NULL) {
-			return NULL;
-		} else {
-			if (what) {
-				memcpy(p, what,
-				       prev_size < size ? prev_size : size);
-			}
-			mm_free(mm, what);
-			return p;
-		}
-	} else {
-		return realloc(what, size);
+			mm->free((void *)what);
 	}
+	else free((void *)what);
 }
 
+/** Realloc implementation using memory context. */
+KR_EXPORT
+void *mm_realloc(knot_mm_t *mm, void *what, size_t size, size_t prev_size);
+
 /** Trivial malloc() wrapper. */
 void *mm_malloc(void *ctx, size_t n);
 
@@ -182,22 +179,56 @@ typedef struct ranked_rr_array_entry ranked_rr_array_entry_t;
 typedef array_t(ranked_rr_array_entry_t *) ranked_rr_array_t;
 /* @endcond */
 
-/** @internal RDATA array maximum size. */
-#define RDATA_ARR_MAX (UINT16_MAX + sizeof(uint64_t))
-
 /** Concatenate N strings. */
 KR_EXPORT
 char* kr_strcatdup(unsigned n, ...);
 
-/** Reseed CSPRNG context. */
-int kr_rand_reseed(void);
-
-/** Get pseudo-random value between zero and max-1 (inclusive).
- *
- * Passing zero means that any uint32_t should be returned (it's also faster).
- */
+/** You probably want kr_rand_* convenience functions instead.
+ * This is a buffered version of gnutls_rnd(GNUTLS_RND_NONCE, ..) */
 KR_EXPORT
-uint32_t kr_rand_uint(uint32_t max);
+void kr_rnd_buffered(void *data, unsigned int size);
+
+/** Return a few random bytes. */
+static inline uint64_t kr_rand_bytes(unsigned int size)
+{
+	uint64_t result;
+	if (size <= 0 || size > sizeof(result)) {
+		kr_log_error("kr_rand_bytes(): EINVAL\n");
+		abort();
+	}
+	uint8_t data[sizeof(result)];
+	kr_rnd_buffered(data, size);
+	/* I would have liked to dump the random data into a size_t directly,
+	 * but that would work well only on little-endian machines,
+	 * so intsead I hope that the compiler will optimize this out.
+	 * (Tested via reading assembly from usual gcc -O2 setup.)
+	 * Alternatively we could waste more rnd bytes, but that seemed worse. */
+	result = 0;
+	for (unsigned int i = 0; i < size; ++ i) {
+		result |= ((size_t)data[i]) << (i * 8);
+	}
+	return result;
+}
+
+/** Throw a pseudo-random coin, succeeding approximately with probability nomin/denomin.
+ * - low precision, only one byte of randomness (or none with extreme parameters)
+ * - tip: use !kr_rand_coin() to get the complementary probability
+ */
+static inline bool kr_rand_coin(unsigned int nomin, unsigned int denomin)
+{
+	/* This function might be called with non-constant values
+	 * so we try to handle odd corner cases instead of crash. */
+	if (nomin >= denomin)
+		return true;
+	else if (nomin <= 0)
+		return false;
+
+	/* threshold = how many parts from 256 are a success */
+	unsigned int threshold = (nomin * 256 + /*rounding*/ denomin / 2) / denomin;
+	if (threshold == 0) threshold = 1;
+	if (threshold == 256) threshold = 255;
+	return (kr_rand_bytes(1) < threshold);
+}
 
 /** Memory reservation routine for knot_mm_t */
 KR_EXPORT
@@ -247,6 +278,9 @@ int kr_sockaddr_cmp(const struct sockaddr *left, const struct sockaddr *right);
 /** Port. */
 KR_EXPORT KR_PURE
 uint16_t kr_inaddr_port(const struct sockaddr *addr);
+/** Set port. */
+KR_EXPORT
+void kr_inaddr_set_port(struct sockaddr *addr, uint16_t port);
 /** String representation for given address as "<addr>#<port>" */
 KR_EXPORT
 int kr_inaddr_str(const struct sockaddr *addr, char *buf, size_t *buflen);
@@ -390,7 +424,7 @@ static inline char *kr_straddr(const struct sockaddr *addr)
 {
 	assert(addr != NULL);
 	/* We are the sinle-threaded application */
-	static char str[INET6_ADDRSTRLEN + 6];
+	static char str[INET6_ADDRSTRLEN + 1 + 5 + 1];
 	size_t len = sizeof(str);
 	int ret = kr_inaddr_str(addr, str, &len);
 	return ret != kr_ok() || len == 0 ? NULL : str;
@@ -435,7 +469,7 @@ static inline int kr_dname_lf(uint8_t *dst, const knot_dname_t *src, bool add_wi
 	}
 	dst[0] = len;
 	return KNOT_EOK;
-};
+}
 
 /* Trivial non-inline wrappers, to be used in lua. */
 KR_EXPORT void kr_rrset_init(knot_rrset_t *rrset, knot_dname_t *owner,
