@@ -10,6 +10,7 @@ import subprocess
 import tempfile
 import time
 
+import dpkt
 import jinja2
 
 from pydnstest import scenario, testserver
@@ -21,7 +22,11 @@ INSTALLDIR = os.path.dirname(os.path.abspath(__file__))
 TRUST_ANCHOR_SUBDIR = 'ta'
 
 
-class IfaceManager(object):
+class DeckardUnderLoadError(Exception):
+    pass
+
+
+class IfaceManager:
     """
     Network interface allocation manager
 
@@ -245,7 +250,7 @@ def run_daemon(cfg, environ):
     logging.getLogger('deckard.daemon.%s.argv' % name).debug('%s', args)
     try:
         proc = subprocess.Popen(args, stdout=daemon_log_file, stderr=subprocess.STDOUT,
-                                cwd=cfg['dir'], preexec_fn=os.setsid, env=environ)
+                                cwd=cfg['dir'], env=environ, start_new_session=True)
     except subprocess.CalledProcessError:
         logger = logging.getLogger('deckard.daemon_log.%s' % name)
         logger.exception("Can't start '%s'", args)
@@ -299,7 +304,7 @@ def process_file(path, qmin, prog_cfgs):
                 'test working directory %s', tmpdir)
         else:
             shutil.rmtree(tmpdir)
-    except:
+    except Exception:
         logging.getLogger('deckard.hint').info(
             'test failed, inspect working directory %s', tmpdir)
         raise
@@ -316,12 +321,27 @@ def setup_daemons(tmpdir, prog_cfgs, template_ctx, ta_files):
         daemons.append({'proc': daemon_proc, 'cfg': prog_cfg})
         try:
             conncheck_daemon(daemon_proc, prog_cfg, template_ctx['_SOCKET_FAMILY'])
-        except:
+        except:  # noqa  -- bare except might be valid here?
             daemon_proc.terminate()
             raise
     return daemons
 
 
+def check_for_icmp():
+    """ Checks Deckards's PCAP for ICMP packets """
+    path = os.environ["SOCKET_WRAPPER_PCAP_FILE"]
+    with open(path, "rb") as f:
+        pcap = dpkt.pcap.Reader(f)
+        for _, packet in pcap:
+            try:
+                ip = dpkt.ip.IP(packet)
+            except dpkt.dpkt.UnpackError:
+                ip = dpkt.ip6.IP6(packet)
+            if isinstance(ip.data, (dpkt.icmp.ICMP, dpkt.icmp6.ICMP6)):
+                return True
+        return False
+
+
 def run_testcase(daemons, case, root_addr, addr_family, prog_under_test_ip):
     """Run actual test and raise exception if the test failed"""
     server = testserver.TestServer(case, root_addr, addr_family)
@@ -344,4 +364,11 @@ def run_testcase(daemons, case, root_addr, addr_family, prog_under_test_ip):
                                  % (daemon['cfg']['name'], daemon['proc'].returncode))
     # Do not clear files if the server crashed (for analysis)
     if server.undefined_answers > 0:
+        # Deckard's responses to resolvers might be delayed due to load which
+        # leads the resolver to close the port and to the test failing in the
+        # end. We partially detect these by checking the PCAP for ICMP packets.
+        if check_for_icmp():
+            logging.error("Deckard is under load.\
+Other errors might be false negatives.\
+Consider retrying the job later.")
         raise ValueError('the scenario does not define all necessary answers (see error log)')
