@@ -1,7 +1,8 @@
-from __future__ import absolute_import
-
+# FIXME pylint: disable=too-many-lines
+from abc import ABC
 import binascii
 import calendar
+from datetime import datetime
 import errno
 import logging
 import os
@@ -11,7 +12,7 @@ import socket
 import string
 import struct
 import time
-from datetime import datetime
+from typing import Optional
 
 import dns.dnssec
 import dns.message
@@ -85,8 +86,10 @@ def sendto_msg(stream, message, addr=None):
             raise
 
 
-def replay_rrs(rrs, nqueries, destination, args=[]):
+def replay_rrs(rrs, nqueries, destination, args=None):
     """ Replay list of queries and report statistics. """
+    if args is None:
+        args = []
     navail, queries = len(rrs), []
     chunksize = 16
     for i in range(nqueries if 'RAND' in args else navail):
@@ -135,6 +138,106 @@ def replay_rrs(rrs, nqueries, destination, args=[]):
     return nsent, nrcvd
 
 
+class DNSBlob(ABC):
+    def to_wire(self) -> bytes:
+        raise NotImplementedError
+
+    def __str__(self) -> str:
+        return '<DNSBlob>'
+
+
+class DNSMessage(DNSBlob):
+    def __init__(self, message: dns.message.Message) -> None:
+        assert message is not None
+        self.message = message
+
+    def to_wire(self) -> bytes:
+        return self.message.to_wire(max_size=65535)
+
+    def __str__(self) -> str:
+        return str(self.message)
+
+
+class DNSReply(DNSMessage):
+    def __init__(
+                self,
+                message: dns.message.Message,
+                query: Optional[dns.message.Message] = None,
+                copy_id: bool = False,
+                copy_query: bool = False
+            ) -> None:
+        super().__init__(message)
+        if copy_id or copy_query:
+            if query is None:
+                raise ValueError("query must be provided to adjust copy_id/copy_query")
+            self.adjust_reply(query, copy_id, copy_query)
+
+    def adjust_reply(
+                self,
+                query: dns.message.Message,
+                copy_id: bool = True,
+                copy_query: bool = True
+            ) -> None:
+        answer = dns.message.from_wire(self.message.to_wire(),
+                                       xfr=self.message.xfr,
+                                       one_rr_per_rrset=True)
+        answer.use_edns(query.edns, query.ednsflags, options=self.message.options)
+        if copy_id:
+            answer.id = query.id
+            # Copy letter-case if the template has QD
+            if answer.question:
+                answer.question[0].name = query.question[0].name
+        if copy_query:
+            answer.question = query.question
+        # Re-set, as the EDNS might have reset the ext-rcode
+        answer.set_rcode(self.message.rcode())
+
+        # sanity check: adjusted answer should be almost the same
+        assert len(answer.answer) == len(self.message.answer)
+        assert len(answer.authority) == len(self.message.authority)
+        assert len(answer.additional) == len(self.message.additional)
+        self.message = answer
+
+
+class DNSReplyRaw(DNSBlob):
+    def __init__(
+                self,
+                wire: bytes,
+                query: Optional[dns.message.Message] = None,
+                copy_id: bool = False
+            ) -> None:
+        assert wire is not None
+        self.wire = wire
+        if copy_id:
+            self.adjust_reply(query, copy_id)
+
+    def adjust_reply(
+                self,
+                query: dns.message.Message,
+                copy_id: bool = True
+            ) -> None:
+        if copy_id:
+            if len(self.wire) < 2:
+                raise ValueError(
+                    'wire data must contain at least 2 bytes to adjust query id')
+            raw_answer = bytearray(self.wire)
+            struct.pack_into('!H', raw_answer, 0, query.id)
+            self.wire = bytes(raw_answer)
+
+    def to_wire(self) -> bytes:
+        return self.wire
+
+    def __str__(self) -> str:
+        return '<DNSReplyRaw>'
+
+
+class DNSReplyServfail(DNSMessage):
+    def __init__(self, query: dns.message.Message) -> None:
+        message = dns.message.make_response(query)
+        message.set_rcode(dns.rcode.SERVFAIL)
+        super().__init__(message)
+
+
 class Entry:
     """
     Data entry represents scripted message and extra metadata,
@@ -155,26 +258,17 @@ class Entry:
         self.fired = 0
 
         # RAW
-        try:
-            self.raw_data = binascii.unhexlify(node["/raw"].value)
-            self.is_raw_data_entry = True
-        except KeyError:
-            self.raw_data = None
-            self.is_raw_data_entry = False
+        self.raw_data = None  # type: Optional[bytes]
+        self.is_raw_data_entry = self.process_raw()
 
         # MATCH
-        self.match_fields = [m.value for m in node.match("/match")]
-
-        if not self.match_fields:
-            self.match_fields = ['opcode', 'qtype', 'qname']
+        self.match_fields = self.process_match()
 
         # FLAGS
-        self.process_reply_line(node)
+        self.process_reply_line()
 
         # ADJUST
-        self.adjust_fields = [m.value for m in node.match("/adjust")]
-        if not self.adjust_fields:
-            self.adjust_fields = ['copy_id']
+        self.adjust_fields = {m.value for m in node.match("/adjust")}
 
         # MANDATORY
         try:
@@ -183,8 +277,53 @@ class Entry:
             self.mandatory = None
 
         # TSIG
+        self.process_tsig()
+
+        # SECTIONS & RECORDS
+        self.sections = self.process_sections()
+
+    def process_raw(self):
         try:
-            tsig = list(node.match("/tsig"))[0]
+            self.raw_data = binascii.unhexlify(self.node["/raw"].value)
+            return True
+        except KeyError:
+            return False
+
+    def process_match(self):
+        try:
+            self.node["/match_present"]
+        except KeyError:
+            return None
+
+        fields = set(m.value for m in self.node.match("/match"))
+
+        if 'all' in fields:
+            fields.remove("all")
+            fields |= set(["opcode", "qtype", "qname", "flags",
+                           "rcode", "answer", "authority", "additional"])
+
+        if 'question' in fields:
+            fields.remove("question")
+            fields |= set(["qtype", "qname"])
+
+        return fields
+
+    def process_reply_line(self):
+        """Extracts flags, rcode and opcode from given node and adjust dns message accordingly"""
+        self.fields = [f.value for f in self.node.match("/reply")]
+        if 'DO' in self.fields:
+            self.message.want_dnssec(True)
+        opcode = self.get_opcode(fields=self.fields)
+        rcode = self.get_rcode(fields=self.fields)
+        self.message.flags = self.get_flags(fields=self.fields)
+        if rcode is not None:
+            self.message.set_rcode(rcode)
+        if opcode is not None:
+            self.message.set_opcode(opcode)
+
+    def process_tsig(self):
+        try:
+            tsig = list(self.node.match("/tsig"))[0]
             tsig_keyname = tsig["/keyname"].value
             tsig_secret = tsig["/secret"].value
             keyring = dns.tsigkeyring.from_text({tsig_keyname: tsig_secret})
@@ -192,11 +331,11 @@ class Entry:
         except (KeyError, IndexError):
             pass
 
-        # SECTIONS & RECORDS
-        self.sections = []
-        for section in node.match("/section/*"):
+    def process_sections(self):
+        sections = set()
+        for section in self.node.match("/section/*"):
             section_name = posixpath.basename(section.path)
-            self.sections.append(section_name)
+            sections.add(section_name)
             for record in section.match("/record"):
                 owner = record['/domain'].value
                 if not owner.endswith("."):
@@ -229,6 +368,7 @@ class Entry:
                     self.message.authority.append(rr)
                 elif section_name == 'additional':
                     self.message.additional.append(rr)
+        return sections
 
     def __str__(self):
         txt = 'ENTRY_BEGIN\n'
@@ -258,19 +398,6 @@ class Entry:
         txt += 'ENTRY_END\n'
         return txt
 
-    def process_reply_line(self, node):
-        """Extracts flags, rcode and opcode from given node and adjust dns message accordingly"""
-        self.fields = [f.value for f in node.match("/reply")]
-        if 'DO' in self.fields:
-            self.message.want_dnssec(True)
-        opcode = self.get_opcode(fields=self.fields)
-        rcode = self.get_rcode(fields=self.fields)
-        self.message.flags = self.get_flags(fields=self.fields)
-        if rcode is not None:
-            self.message.set_rcode(rcode)
-        if opcode is not None:
-            self.message.set_opcode(opcode)
-
     @classmethod
     def get_flags(cls, fields):
         """From `fields` extracts and returns flags"""
@@ -297,7 +424,7 @@ class Entry:
                 pass
         if len(rcodes) > 1:
             raise ValueError("Parse failed, too many rcode values.", rcodes)
-        if len(rcodes) == 0:
+        if not rcodes:
             return None
         return rcodes[0]
 
@@ -315,17 +442,13 @@ class Entry:
                 pass
         if len(opcodes) > 1:
             raise ValueError("Parse failed, too many opcode values.")
-        if len(opcodes) == 0:
+        if not opcodes:
             return None
         return opcodes[0]
 
     def match(self, msg):
         """ Compare scripted reply to given message based on match criteria. """
-        match_fields = self.match_fields
-        if 'all' in match_fields:
-            match_fields.remove('all')
-            match_fields += ['flags'] + ['rcode'] + self.sections
-        for code in match_fields:
+        for code in self.match_fields:
             try:
                 pydnstest.matchpart.match_part(self.message, msg, code)
             except pydnstest.matchpart.DataMismatch as ex:
@@ -344,40 +467,16 @@ class Entry:
         if expected != got:
             raise ValueError("raw message comparsion failed: expected %s got %s" % (expected, got))
 
-    def _adjust_reply(self, query):
-        """ Copy scripted reply and adjust to received query. """
-        answer = dns.message.from_wire(self.message.to_wire(),
-                                       xfr=self.message.xfr,
-                                       one_rr_per_rrset=True)
-        answer.use_edns(query.edns, query.ednsflags, options=self.message.options)
-        if 'copy_id' in self.adjust_fields:
-            answer.id = query.id
-            # Copy letter-case if the template has QD
-            if answer.question:
-                answer.question[0].name = query.question[0].name
-        if 'copy_query' in self.adjust_fields:
-            answer.question = query.question
-        # Re-set, as the EDNS might have reset the ext-rcode
-        answer.set_rcode(self.message.rcode())
-
-        # sanity check: adjusted answer should be almost the same
-        assert len(answer.answer) == len(self.message.answer)
-        assert len(answer.authority) == len(self.message.authority)
-        assert len(answer.additional) == len(self.message.additional)
-        return answer
-
-    def _adjust_raw_reply(self, query):
-        if 'raw_id' in self.adjust_fields:
-            assert len(self.raw_data) >= 2, "RAW message has to contain at least 2 bytes"
-            raw_answer = bytearray(self.raw_data)
-            struct.pack_into('!H', raw_answer, 0, query.id)
-            return bytes(raw_answer)
-        return self.raw_data
-
-    def reply(self, query):
+    def reply(self, query) -> Optional[DNSBlob]:
+        if 'do_not_answer' in self.adjust_fields:
+            return None
         if self.is_raw_data_entry:
-            return self._adjust_raw_reply(query), True
-        return self._adjust_reply(query), False
+            copy_id = 'raw_data' in self.adjust_fields
+            assert self.raw_data is not None
+            return DNSReplyRaw(self.raw_data, query, copy_id)
+        copy_id = 'copy_id' in self.adjust_fields
+        copy_query = 'copy_query' in self.adjust_fields
+        return DNSReply(self.message, query, copy_id, copy_query)
 
     def set_edns(self, fields):
         """ Set EDNS version and bufsize. """
@@ -426,7 +525,7 @@ class Range:
 
         address = node["/address"].value
         self.addresses = {address} if address is not None else set()
-        self.addresses |= set([a.value for a in node.match("/address/*")])
+        self.addresses |= {a.value for a in node.match("/address/*")}
         self.stored = [Entry(n) for n in node.match("/entry")]
         self.args = {}
         self.received = 0
@@ -447,23 +546,16 @@ class Range:
         txt += 'RANGE_END\n\n'
         return txt
 
-    def eligible(self, id, address):
+    def eligible(self, ident, address):
         """ Return true if this range is eligible for fetching reply. """
-        if self.a <= id <= self.b:
+        if self.a <= ident <= self.b:
             return (None is address
                     or set() == self.addresses
                     or address in self.addresses)
         return False
 
-    def reply(self, query):
-        """
-        Get answer for given query (adjusted if needed).
-
-        Returns: (answer, is_raw_data)
-            answer: DNS message object or wire-format (bytes) or None if there is no
-                candidate in this range
-            is_raw_data: True if response is wire-format bytes, instead of DNS message object
-        """
+    def reply(self, query: dns.message.Message) -> Optional[DNSBlob]:
+        """Get answer for given query (adjusted if needed)."""
         self.received += 1
         for candidate in self.stored:
             try:
@@ -472,13 +564,13 @@ class Range:
                 # Probabilistic loss
                 if 'LOSS' in self.args:
                     if random.random() < float(self.args['LOSS']):
-                        return None, None
+                        return DNSReplyServfail(query)
                 self.sent += 1
                 candidate.fired += 1
                 return resp
             except ValueError:
                 pass
-        return None, None
+        return DNSReplyServfail(query)
 
 
 class StepLogger(logging.LoggerAdapter):  # pylint: disable=too-few-public-methods
@@ -557,9 +649,9 @@ class Step:
             # Parse QUERY-specific parameters
             choice, tcp, source = None, False, None
             return self.__query(ctx, tcp=tcp, choice=choice, source=source)
-        elif self.type == 'CHECK_OUT_QUERY':
+        elif self.type == 'CHECK_OUT_QUERY':  # ignore
             self.log.info('')
-            pass  # Ignore
+            return None
         elif self.type == 'CHECK_ANSWER' or self.type == 'ANSWER':
             self.log.info('')
             return self.__check_answer(ctx)
@@ -568,6 +660,7 @@ class Step:
             return self.__time_passes()
         elif self.type == 'REPLY' or self.type == 'MOCK':
             self.log.info('')
+            return None
         # Parser currently doesn't support step types LOG, REPLAY and ASSERT.
         # No test uses them.
         # elif self.type == 'LOG':
@@ -728,25 +821,19 @@ class Scenario:
         if self.info:
             txt += ' {0}'.format(self.info)
         txt += '\n'
-        for range in self.ranges:
-            txt += str(range)
+        for range_ in self.ranges:
+            txt += str(range_)
         for step in self.steps:
             txt += str(step)
         txt += "\nSCENARIO_END"
         return txt
 
-    def reply(self, query, address=None):
-        """
-        Generate answer packet for given query.
-
-        The answer can be DNS message object or a binary blob.
-        Returns:
-            (answer, boolean "is the answer binary blob?")
-        """
+    def reply(self, query: dns.message.Message, address=None) -> Optional[DNSBlob]:
+        """Generate answer packet for given query."""
         current_step_id = self.current_step.id
         # Unknown address, select any match
         # TODO: workaround until the server supports stub zones
-        all_addresses = set()
+        all_addresses = set()  # type: ignore
         for rng in self.ranges:
             all_addresses.update(rng.addresses)
         if address not in all_addresses:
@@ -767,7 +854,7 @@ class Scenario:
                 return candidate.reply(query)
             except (IndexError, ValueError):
                 pass
-        return None, True
+        return DNSReplyServfail(query)
 
     def play(self, paddr):
         """ Play given scenario. """
@@ -819,15 +906,15 @@ def get_next(file_in, skip_empty=True):
         if not line:
             return False
         quoted, escaped = False, False
-        for i in range(len(line)):
-            if line[i] == '\\':
+        for i, char in enumerate(line):
+            if char == '\\':
                 escaped = not escaped
-            if not escaped and line[i] == '"':
+            if not escaped and char == '"':
                 quoted = not quoted
-            if line[i] in ';' and not quoted:
+            if char == ';' and not quoted:
                 line = line[0:i]
                 break
-            if line[i] != '\\':
+            if char != '\\':
                 escaped = False
         tokens = ' '.join(line.strip().split()).split()
         if not tokens:
@@ -839,7 +926,7 @@ def get_next(file_in, skip_empty=True):
         return op, tokens
 
 
-def parse_config(scn_cfg, qmin, installdir):
+def parse_config(scn_cfg, qmin, installdir):  # FIXME: pylint: disable=too-many-statements
     """
     Transform scene config (key, value) pairs into dict filled with defaults.
     Returns tuple:
